#!/usr/bin/python
# code to convert a tempalte svn authZ file with LIGO.ORG grouper groups to a
# useable final authZ file for svn.  It expands the group entries to all the
# members of the group.
#
# This should be run periodically to pick up any changes.
#
# Jonathan Hanks LIGO Hanford Observatory 2009
#
# modified in 2015 by Jonathan Hanks to be used for nds2 server user acl files.
import ConfigParser
import sys
import ldap
import sys
import re

class settingsDB:
    """settingsDB is a dictionary like structure that holds configuration information.
    
    This could just be a dictionary, but I had stolen this code from a project that actually initialized itself from disk,
    hence the new class.  Maybe someday this will read from disk."""
    def __init__(self, fname):
        settings = {}
        if fname != '':
            cfg = ConfigParser.ConfigParser()
            cfg.read(fname)

            stn = 'ldap'
            settings['ldap.primaryServer'] = cfg.get(stn, 'server')
            settings['ldap.secondaryServer'] = ''
            settings['ldap.port'] = cfg.getint(stn, 'port')
            settings['ldap.ssl'] = cfg.getboolean(stn, 'ssl')
            settings['ldap.startTLS'] = cfg.getboolean(stn, 'starttls')

            stn = 'users'
            settings['ldap.groupBase'] = cfg.get(stn, 'base')
            settings['ldap.searchGroupMembers'] = cfg.get(stn, 'filter')
            settings['ldap.searchGroupMembersAttr'] = [cfg.get(stn, 'attr'),]

            stn = 'groups'
            settings['ldap.listBase'] = cfg.get(stn, 'base')
            settings['ldap.searchListGroups'] = cfg.get(stn, 'filter')
            settings['ldap.searchListGroupsAttr'] = [cfg.get(stn, 'attr'),]
        else:
            settings['ldap.primaryServer'] = 'quartz.ligo-wa.caltech.edu'
            settings['ldap.secondaryServer'] = ''
            settings['ldap.port'] = 636
            settings['ldap.ssl'] = True
            settings['ldap.startTLS'] = False
            settings['ldap.groupBase'] = 'dc=ligo,dc=org'
            settings['ldap.searchGroupMembers'] = '(&(objectClass=inetOrgPerson)(isMemberOf=%s))'
            settings['ldap.searchGroupMembersAttr'] = ['krbPrincipalName',]
            settings['ldap.listBase'] = 'ou=Communities,ou=grouper,dc=ligo,dc=org'
            settings['ldap.searchListGroups'] = '(objectClass=groupOfNames)'
            settings['ldap.searchListGroupsAttr'] = ['description',]

        self._settings = settings
    
    def __len__(self):
        return len(self._settings)
    
    def __getitem__(self,key):
        if (self._settings.has_key(key)):
            return self._settings[key]
        raise IndexError('Unknown setting key %s' % key)
    
    def __iter__(self):
        return self._settings.iterkeys()
    
    def iterkeys(self):
        return self._settings.iterkeys()

    def keys(self):
        return self._settings.keys()
    
    def has_key(self, key):
        return self._settings.has_key(key)
    
    def __contains(self, key):
        return self.has_key(key)

    def dump(self, f):
        cfg = ConfigParser.RawConfigParser()

        stn = 'ldap'
        cfg.add_section(stn)
        cfg.set(stn, 'starttls', str(self._settings['ldap.startTLS']))
        cfg.set(stn, 'ssl', str(self._settings['ldap.ssl']))
        cfg.set(stn, 'port', self._settings['ldap.port'])
        cfg.set(stn, 'server', self._settings['ldap.primaryServer'])

        stn = 'users'
        cfg.add_section(stn)
        cfg.set(stn, 'attr', self._settings['ldap.searchGroupMembersAttr'][0])
        cfg.set(stn, 'filter', self._settings['ldap.searchGroupMembers'])
        cfg.set(stn, 'base', self._settings['ldap.groupBase'])

        stn = 'groups'
        cfg.add_section(stn)
        cfg.set(stn, 'attr', self._settings['ldap.searchListGroupsAttr'][0])
        cfg.set(stn, 'filter', self._settings['ldap.searchListGroups'])
        cfg.set(stn, 'base', self._settings['ldap.listBase'])

        cfg.write(f)

class grouper:
    """not really grouper, but it would be if I had access to it"""
    
    def __init__(self, settings):
        """Initialize the object pulling parameters from the settings object."""
        self._primaryServer = settings['ldap.primaryServer']
        self._secondaryServer = settings['ldap.secondaryServer']
        self._port = settings['ldap.port']
        self._ssl = settings['ldap.ssl']
        self._startTLS = settings['ldap.startTLS']
        self._groupBase = settings['ldap.groupBase']
        self._searchGroupMembers = settings['ldap.searchGroupMembers']
        self._searchGroupMembersAttr = settings['ldap.searchGroupMembersAttr']
        self._listBase = settings['ldap.listBase']
        self._searchListGroups = settings['ldap.searchListGroups']
        self._searchListGroupsAttr = settings['ldap.searchListGroupsAttr']
        self._connection = None
        self._groups = {}

    def __del__(self):
        """cleanup, close the ldap connection if it is open"""
        if (self._connection != None):
            self._connection.unbind_s()
            self._connection = None
            #print 'Connection to LDAP destroyed'

    def connect(self):
        """connect the the ldap."""
        if self._connection != None:
            # already connected
            return
        secured = ''
        if (self._ssl == True):
            secured = 's'
        self._connection = ldap.initialize("ldap%s://%s:%d" % (secured, self._primaryServer, self._port))
        #print 'Connected to LDAP'
        if (self._startTLS):
            self._connection.start_tls_s()
            #print 'STARTTLS'
        self._connection.simple_bind_s()
        #print 'Bind completed'
        
    def getGroups(self):
        """Retreive a list of all grouper groups from the ldap.  The current iteration holds the list in-memory."""
        if (self._connection == None):
            self.connect()
        groupList = {}
        query = self._searchListGroups
        r = self._connection.search_s(self._listBase, ldap.SCOPE_SUBTREE, query, self._searchListGroupsAttr)
        for entry in r:
            dn = entry[0]
            contents = entry[1]
            if (len(contents['description']) > 0):
                groupList[contents['description'][0]] = 1
        keys = groupList.keys()
        keys.sort()
        return keys
        
    def getGroupMembers(self, groupName):
        """Given a group group name return a list of members.  The current iteration holds the list in-memory."""
        if (self._groups.has_key(groupName)):
            return self._groups[groupName]
        if (self._connection == None):
            self.connect()
        members = []
        query = self._searchGroupMembers % groupName
        r = self._connection.search_s(self._groupBase,ldap.SCOPE_SUBTREE, query, self._searchGroupMembersAttr)
        #print "self._connection.search_s('%s', ldap.SCOPE_SUBTREE, '%s', '%s')" % (self._groupBase, query, self._searchGroupMembersAttr)
        #print 'Found %d results' % len(r)
        for entry in r:
            #print entry
            dn = entry[0]
            contents = entry[1]
            #print contents
            if (len(contents['krbPrincipalName']) > 0):
                members.append(contents['krbPrincipalName'][0])
            self._groups[groupName] = members
        return members

def processAccessList(groups, fin, fout):
    """Given a svn authZ file (fin) process it expanding grouper groups.  Output is sent to the file object fout."""
    blacklist_re = re.compile("^-([0-9a-zA-Z_:\-\.\\/]+)$")
    prog_re = re.compile("^@([0-9a-zA-Z_:\-\.]+)(::[0-9])$")
    blacklisted = {}
    for line in fin:
        # check for black list definitions
        bl_result = blacklist_re.match(line)
        if bl_result != None:
            bl_name = bl_result.group(1)
            fout.write("# blacklisting -" + bl_name+"\n")
            blacklisted[bl_name] = True
        else:
            result = prog_re.match(line)
            if (result == None):
                fout.write(line)
            else:
                fout.write("#" + line)
                groupName = result.group(1)
                rights = result.group(2)
                members = groups.getGroupMembers(groupName)
                for member in members:
                    clean_name = member.split('@', 1)[0]
                    if blacklisted.has_key(clean_name):
                        fout.write("#" + clean_name + rights + "\n")
                    else:
                        fout.write(clean_name + rights + "\n")
                fout.write("#end of group match\n")

def Useage():
    print "Useage:\n\t%s <options> <infile> <outfile>\n" % sys.argv[0]
    print "Options:\n-h this help\n-l list available groups\n"
    print "-c <configfile>\n-dc\n"
    print "\nThe -dc option dumps a sample config file to stdout and quits"
    print ""
    print "infile - nds2 userfile.txt file (defaults to stdin)"
    print "outfile - nds2 userfile.txt file with group references expanded using ldap and grouper groups (defaults to stdout)"
    sys.exit(1)
    
def listGroups(groups):
    groupList = groups.getGroups()
    for group in groupList:
        print group

fin = sys.stdin
fout = sys.stdout

assignConfig = False
finAssigned = False
foutAssigned = False
doList = False
doAuthz = True
cfgFile = ""
for i in range(1, len(sys.argv)):
    #print 'Parsing sys.argv[i] = %s' % sys.argv[i]
    if (sys.argv[i] == '-h'):
        Useage()
    elif (sys.argv[i] == '-l'):
        doList = True
        doAuthz = False
    elif (sys.argv[i] == '-c'):
        assignConfig = True
    elif (sys.argv[i] == '-dc'):
        settings = settingsDB('')
        settings.dump(sys.stdout)
        sys.exit(0)
    elif assignConfig:
        assignConfig = False
        cfgFile = sys.argv[i]
    elif (not finAssigned):
        finAssigned = True
        if (sys.argv[i] != '-'):
            fin = open(sys.argv[i], "rt")
    elif (not foutAssigned):
        foutAssigned = True
        if (sys.argv[i] != '-'):
            fout = open(sys.argv[i], "wt")
    else:
        Useage()

settings = settingsDB(cfgFile)
groups = grouper(settings)

if (doList):
    listGroups(groups)
if (doAuthz):
    processAccessList(groups, fin, fout)

fout.flush()
fout.close()
sys.exit(0)

